#
# $Id: jncalculator.y,v 1.8 2005/08/22 09:13:51 aamine Exp $
#
# Japanese Numeric Calculator
#

class JNCalculator
  options no_result_var

  prechigh
    nonassoc UMINUS
    left '*' '/'
    left '+' '-'
  preclow
rule
  stmt    : expr
          | /* empty */      { nil }

  expr    : expr '+' expr    { val[0] + val[2] }
          | expr '-' expr    { val[0] - val[2] }
          | expr '*' expr    { val[0] * val[2] }
          | expr '/' expr    { val[0] / val[2] }
          | '(' expr ')'     { val[1] }
          | NUM
          | '-' NUM  =UMINUS { -(val[1]) }
end

---- header
#! ruby -Ke

require 'rational'
require 'strscan'

SPACES = /[\s]+/

---- inner

  def initialize(numdef)
    @numdef = numdef
  end

  def parse(str)
    @source = str
    yyparse(self, :yylex)
  end

  private

  def yylex
    s = StringScanner.new(@source)
    until s.eos?
      case
      when s.scan(::SPACES)
        ;
      when tok = s.scan(/[0-9-]+/)
        yield :NUM, Rational(tok.to_i)
      when tok = s.scan(@numdef.chars_re)
        yield :NUM, Rational(@numdef.parse(tok))
      when tok = s.scan(/[\+\-\*\/\(\)ܡݡߡʡ]/)
        yield to_rubyop(tok), tok
      else
        raise ParseError, "parse error on char #{s.getch.inspect}"
      end
    end
    yield nil
  end

  def to_rubyop(tok)
    { '' => '+',
      '' => '-',
      '' => '*',
      '' => '/',
      '' => '(',
      '' => ')' }[tok] || tok
  end

---- footer

require 'component'

class String
  def mbstrip
    sub(/\A#{::SPACES}/o, '').sub(/#{::SPACES}\z/o, '')
  end
end

def jncalculator_main
  while line = prompt('JCALC> ')
    exit if /q(u(it?)?)?/i =~ line.mbstrip
    next if line.mbstrip.empty?
    begin
      puts lldn_eval(line.mbstrip)
    rescue => err
      puts "Error: #{err.message}"
    end
  end
end

def lldn_eval(str)
  @parser ||= JNCalculator.new(Component.LLDN)
  Component.LLDN.string_expr(@parser.parse(str.mbstrip))
end

begin
  require 'readline'
  def prompt(str)
    Readline.readline(str)
  end
rescue LoadError
  def prompt(str)
    print str; $stdout.flush
    return $stdin.gets
  end
end

if $0 == __FILE__
  jncalculator_main
end
